/* ====================================================================
 * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/* 
 * mod_auth_cert.cpp - calls the Akenti policy engine to determine
 *         if a user has access. Assumes user has connected via
 *          https and has presented a x509 indentity certificate
 * Modified Apache example module. Things that we do not do use the
 *   original example module code.
 *
 *  Sept 12, 1998 - mrt
 *     updated for apache 1.3.1 API 
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include <stdio.h>

/* get around #define conflicts -- used instead of enumerated types */
#define H_OK 0 /* OK */
#undef OK

/* Security library includes */
#include "Certificate_X509.h"
#include "AuthCertModule.h"
#include "AkentiConf.h"
#include "UseCond.h"
#include "ListUseCond.h"
#include "Policy.h"
#include "ListPolicy.h"
#include "RootPolicy.h"
#include "RootPolicy_util.h"
#include "DirectoryServiceUtil.h"
#include "CertificateQueryHandler_attribute_text.h"
#include "CertificateQueryHandler_X509.h"
#include "CertificateRetrieval.h"
#include "UseCond_util.h"
#include "BasicString_util.h"
#include "Certificate_util.h"
#include "Certificate_text.h"
#include "MIME_util.h"
#include "C_RSA_utils.h"
#include "baselib.h"
#include <iostream.h>
#include <stdlib.h>
#include "File_util.h"
#include "AKList.h"
#include "Search.h"

#include "WebDirectoryServiceImp.h"
#include "FileDirectoryServiceImp.h"
#include "LDAPDirectoryServiceImp.h"


/*--------------------------------------------------------------------------*/
/*									    */
/* Data declarations.							    */
/*									    */
/* Here are the static cells and structure declarations private to our	    */
/* module.								    */
/*									    */
/*--------------------------------------------------------------------------*/

/*
 * Sample configuration record.  Used for both per-directory and per-server
 * configuration data.
 *
 * It's perfectly reasonable to have two different structures for the two
 * different environments.  The same command handlers will be called for
 * both, though, so the handlers need to be able to tell them apart.  One
 * possibility is for both structures to start with an int which is zero for
 * one and 1 for the other.
 */
typedef struct auth_cert_config {
    int	    cmode;	/* Environment to which record applies (directory,  */
			/* server, or combination).			    */
#define CONFIG_MODE_SERVER 1
#define CONFIG_MODE_DIRECTORY 2
#define CONFIG_MODE_COMBO 3  /* Shouldn't ever happen.			    */
    int	    local;	/* Boolean: was "Example" directive declared here?  */
    int	    congenital;	/* Boolean: did we inherit an "Example"?	    */
    char    *trace;	/* Pointer to trace string.			    */
    BasicString accessfile;/* name of the htaccess file to be parsed	    */
    AkentiConf *config;  /* pathname of the Akenti configuration file      */
    char    *loc;	/* Location to which this record applies.	    */
} auth_cert_config;

static auth_cert_config *our_dconfig(request_rec *r);
static auth_cert_config *our_sconfig(server_rec *s);

/*
 * Let's set up a module-local static cell to point to the accreting callback
 * trace.  As each API callback is made to us, we'll tack on the particulars
 * to whatever we've already recorded.
 */
static char *trace = NULL;

/*
 * To avoid leaking memory from pools other than the per-request one, we
 * allocate a module-private pool, and then use a sub-pool of that which gets
 * freed each time we modify the trace.  That way previous layers of trace
 * data don't get lost.
 */
static pool *example_pool = NULL;
static pool *example_subpool = NULL;

/*
 * Declare ourselves so the configuration routines can find and know us.
 * We'll fill it in at the end of the module.
 */
/* module auth_cert_module;*/

/*--------------------------------------------------------------------------*/
/*									    */
/* The following pseudo-prototype declarations illustrate the parameters    */
/* passed to command handlers for the different types of directive	    */
/* syntax.  If an argument was specified in the directive definition	    */
/* (look for "command_rec" below), it's available to the command handler    */
/* via the (void *) info field in the cmd_parms argument passed to the	    */
/* handler (cmd->info for the examples below).				    */
/*									    */
/*--------------------------------------------------------------------------*/

/*
 * Command handler for a NO_ARGS directive.
 *
 * static const char *handle_NO_ARGS
 *	(cmd_parms *cmd, void *mconfig);
 */
 
/*
 * Command handler for a RAW_ARGS directive.  The "args" argument is the text
 * of the commandline following the directive itself.
 *
 * static const char *handle_RAW_ARGS
 *	(cmd_parms *cmd, void *mconfig, const char *args);
 */

/*
 * Command handler for a TAKE1 directive.  The single parameter is passed in
 * "word1".
 *
 * static const char *handle_TAKE1
 *	(cmd_parms *cmd, void *mconfig, char *word1);
 */

/*
 * Command handler for a TAKE2 directive.  TAKE2 commands must always have
 * exactly two arguments.
 *
 * static const char *handle_TAKE2
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2);
 */

/*
 * Command handler for a TAKE3 directive.  Like TAKE2, these must have exactly
 * three arguments, or the parser complains and doesn't bother calling us.
 *
 * static const char *handle_TAKE3
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2, char *word3);
 */

/*
 * Command handler for a TAKE12 directive.  These can take either one or two
 * arguments.
 * - word2 is a NULL pointer if no second argument was specified.
 *
 * static const char *handle_TAKE12
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2);
 */

/*
 * Command handler for a TAKE123 directive.  A TAKE123 directive can be given,
 * as might be expected, one, two, or three arguments.
 * - word2 is a NULL pointer if no second argument was specified.
 * - word3 is a NULL pointer if no third argument was specified.
 *
 * static const char *handle_TAKE123
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2, char *word3);
 */

/*
 * Command handler for a TAKE13 directive.  Either one or three arguments are
 * permitted - no two-parameters-only syntax is allowed.
 * - word2 and word3 are NULL pointers if only one argument was specified.
 *
 * static const char *handle_TAKE13
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2, char *word3);
 */

/*
 * Command handler for a TAKE23 directive.  At least two and as many as three
 * arguments must be specified.
 * - word3 is a NULL pointer if no third argument was specified.
 *
 * static const char *handle_TAKE23
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2, char *word3);
 */

/*
 * Command handler for a ITERATE directive.
 * - Handler is called once for each of n arguments given to the directive.
 * - word1 points to each argument in turn.
 *
 * static const char *handle_ITERATE
 *	(cmd_parms *cmd, void *mconfig, char *word1);
 */

/*
 * Command handler for a ITERATE2 directive.
 * - Handler is called once for each of the second and subsequent arguments
 *   given to the directive.
 * - word1 is the same for each call for a particular directive instance (the
 *   first argument).
 * - word2 points to each of the second and subsequent arguments in turn.
 *
 * static const char *handle_ITERATE2
 *	(cmd_parms *cmd, void *mconfig, char *word1, char *word2);
 */

/*
 * This routine is used to add a trace of a callback to the list.  We're
 * passed the server record (if available), an allocation pool, a pointer to
 * our private configuration record (if available) for the environment to
 * which the callback is supposed to apply, and some text.  We turn this into
 * a textual representation and add it to the tail of the list.  The list can
 * be displayed by the auth_cert_handler() routine.
 */
static void trace_add
	(server_rec *s, auth_cert_config *mconfig, const char *note) {

    char    *sofar;
    char    *addon;
    char    *where;
    pool    *subpool;

    /*
     * Make a new sub-pool and copy any existing trace to it.
     */
    subpool = ap_make_sub_pool (example_pool);
    if (trace != NULL) {
	addon = ap_pstrcat (subpool, trace, NULL);
    }
    /*
     * Now, if we have a sub-pool from before, nuke it and replace with the
     * one we just allocated.
     */
    if (example_subpool != NULL) {
	ap_destroy_pool (example_subpool);
    }
    example_subpool = subpool;
    /*
     * If we weren't passed a configuration record, we can't figure out to
     * what location this call applies.  This only happens for co-routines
     * that don't operate in a particular directory or server context.  If we
     * got a valid record, extract the location (directory or server) to which
     * it applies.
     */
    where = (mconfig != NULL) ? mconfig->loc : "nowhere";
    where = (where != NULL) ? where : "";
    addon = ap_pstrcat 
		(
		    subpool,
		    "   <LI>\n",
		    "    <DL>\n",
		    "     <DT><SAMP>",
		    note,
		    "</SAMP>\n",
		    "     </DT>\n",
		    "     <DD><SAMP>[",
		    where,
		    "]</SAMP>\n",
		    "     </DD>\n",
		    "    </DL>\n",
		    "   </LI>\n",
		    NULL
		);
    sofar = (trace == NULL) ? "" : trace;
    trace = ap_pstrcat (subpool, sofar, addon, NULL);
    /*
     * Store a copy of the same information in the configuration record, if
     * there is one.
     */
    if (mconfig != NULL) {
	sofar = (mconfig->trace == NULL) ? "" : mconfig->trace;
	mconfig->trace = ap_pstrcat (subpool, sofar, addon, NULL);
    }
    /*
     * You *could* uncomment the following if you wanted to see the calling
     * sequence reported in the server's error_log, but beware - almost all of
     * these co-routines are called for every single request, and the impact
     * on the size (and readability) of the error_log is considerable.
     */
/*
    if (s != NULL) {
        log_printf (s, "mod_example: %s", note);
    }
 */
}

/*--------------------------------------------------------------------------*/
/*									    */
/* We prototyped the various syntax for command handlers (routines that     */
/* are called when the configuration parser detects a directive declared    */
/* by our module) earlier.  Now we actually declare a "real" routine that   */
/* will be invoked by the parser when our "real" directive is		    */
/* encountered.								    */
/*									    */
/* If a command handler encounters a problem processing the directive, it   */
/* signals this fact by returning a non-NULL pointer to a string	    */
/* describing the problem.						    */
/*--------------------------------------------------------------------------*/
/* 
 * Command handler for the NO_ARGS "Example" directive.  All we do is mark the
 * call in the trace log, and flag the applicability of the directive to the
 * current location in that location's configuration record.
 */
/* Debugging routine */
void show_env(request_rec *r){
  char *remotehost = getenv("REMOTE_HOST");
  char *akenticonf = getenv("AKENTI_CONF");
  remotehost = r->connection->remote_ip;
  akenticonf = ap_table_get(r->subprocess_env, "akenti_conf");
}

/* 
 * A common function which returns the last component of a pathname
 */
void gethtaccess(BasicString &in, BasicString *out) {
	BasicString_util bs;

	BasicString t1,t2;
	t1 = in;
	while (bs.after(t1, "/", &t2) == BasicString_util::OK) {
		t1 = t2;
	 }
	*out = t1;
 }

/* process AuthenticateWithCert directive */
static const char *cmd_auth_cert
	(cmd_parms *cmd, void *mconfig) {

    auth_cert_config
	    *cfg = (auth_cert_config *) mconfig;

	if(cfg->accessfile.rep() == NULL) {
		BasicString out;
		gethtaccess(BasicString(cmd->config_file->name), &out);
		cfg->accessfile = out;
                //cout << "The htaccess name : " << cfg->accessfile.rep() << endl;
	  }

    /*
     * "Example Wuz Here"
     */
    cfg->local = 1;
    trace_add (cmd->server, cfg, "cmd_example()");
    return NULL;
}

/*
 * @author : Srilekha S. Mudumbai
 * Last Update : October 6, 1997
 */

const char *cmd_auth_certauthority
	(cmd_parms *cmd, void *mconfig, const char *dn, const char *cert) {
	auth_cert_config
		*cfg = (auth_cert_config *) mconfig;

	if(cfg->accessfile.rep() == NULL) {
		BasicString out;
		gethtaccess(BasicString(cmd->config_file->name), &out);
		cfg->accessfile = out;
                //cout << "The htaccess name : " << cfg->accessfile.rep() << endl;
	  }

	cfg->local=1;
	return NULL;
 }

/* 
 * Modifying this module: Srilekha S Mudumbai
 * Last Update : July 24, 1997
 */

const char *cmd_auth_usecond
	(cmd_parms *cmd, void *mconfig, const char *url) {
    	auth_cert_config
       	     *cfg = (auth_cert_config *) mconfig;

	if(cfg->accessfile.rep() == NULL) {
		BasicString out;
		gethtaccess(BasicString(cmd->config_file->name), &out);
		cfg->accessfile = out;
                //COUt << "The htaccess name : " << cfg->accessfile.rep() << endl;
	  }
    /*
     * "Example Wuz Here"
     */
    cfg->local = 1;
    trace_add (cmd->server, cfg, "cmd_example()");
    return NULL;
}

/* found the AkentiConf directive 
 *   read the configuration file and initalize the AkentiConf structure 
 *   Save the configuration file name in an environment variable for use
 *   by scripts
 */
const char *cmd_auth_akenticonf
	(cmd_parms *cmd, void *mconfig, const char *akenticonfname) {
    	auth_cert_config
       	     *cfg = (auth_cert_config *) mconfig;
        /* table *e = r->subprocess_env; */

	if (cfg->config == NULL) {
	  cfg->config = new AkentiConf;
	}
	if (cfg->config->InitializeFromFile(akenticonfname) == AkentiConf::FAIL){
	  cerr << "unable to read Akenti Configuration File: '"
	       << akenticonfname << endl;
	}
	/* ap_table_set(e,"AKENTI_CONF",akenticonfname); */

	cfg->local = 1;
	trace_add (cmd->server, cfg, "cmd_example()");
	
	auth_cert_config *serv_cfg = our_sconfig(cmd->server);
	if (serv_cfg->config == NULL) {
	  serv_cfg->config = cfg->config;
	}
	return NULL;
}

/*
 * Module : _usecondissuer
 * @author : Srilekha Mudumbai
 * Updated : Oct 31, 1997
 */
const char *cmd_auth_usecondissuer
	(cmd_parms *cmd, void *mconfig, const char *dnca) {
    	auth_cert_config
       	     *cfg = (auth_cert_config *) mconfig;

	if(cfg->accessfile.rep() == NULL) {
		BasicString out;
		gethtaccess(BasicString(cmd->config_file->name), &out);
		cfg->accessfile = out;
                //cout << "The htaccess name : " << cfg->accessfile.rep() << endl;
	  }
    /*
     * "Example Wuz Here"
     */
    cfg->local = 1;
    trace_add (cmd->server, cfg, "cmd_example()");
    return NULL;
}

const char *cmd_auth_useconddir
	(cmd_parms *cmd, void *mconfig, const char *trusted,
	 const char *type, const char *dirarg) {

	auth_cert_config *cfg = (auth_cert_config *)mconfig;

	if(cfg->accessfile.rep() == NULL) {
		BasicString out;
		gethtaccess(BasicString(cmd->config_file->name), &out);
		cfg->accessfile = out;
                //cout << "The htaccess name : " << cfg->accessfile.rep() << endl;
	  }

	cfg->local = 1;
        trace_add (cmd->server, cfg, "cmd_example()");
	return NULL;
 }

const char *cmd_auth_certdir
	(cmd_parms *cmd, void *mconfig, const char *trusted,
	 const char *type, const char *dirarg) {

    auth_cert_config
	    *cfg = (auth_cert_config *) mconfig;

	if(cfg->accessfile.rep() == NULL) {
		BasicString out;
		gethtaccess(BasicString(cmd->config_file->name), &out);
		cfg->accessfile = out;
                //cout << "The htaccess name : " << cfg->accessfile.rep() << endl;
	  }
    /*
     * "Example Wuz Here"
     */
    cfg->local = 1;
    trace_add (cmd->server, cfg, "cmd_example()");
    return NULL;
}

/*--------------------------------------------------------------------------*/
/*									    */
/* Now we declare our content handlers, which are invoked when the server   */
/* encounters a document which our module is supposed to have a chance to   */
/* see.  (See mod_mime's SetHandler and AddHandler directives, and the	    */
/* mod_info and mod_status examples, for more details.)			    */
/*									    */
/* Since content handlers are dumping data directly into the connexion	    */
/* (using the r*() routines, such as ap_rputs() and ap_rprintf()) without	    */
/* intervention by other parts of the server, they need to make		    */
/* sure any accumulated HTTP headers are sent first.  This is done by	    */
/* calling ap_set_http_header().  Otherwise, no header will be sent at all,   */
/* and the output sent to the client will actually be HTTP-uncompliant.	    */
/*--------------------------------------------------------------------------*/
/* 
 * Sample content handler.  All this does is display the call list that has
 * been built up so far.
 *
 * The return value instructs the caller concerning what happened and what to
 * do next:
 *  OK ("we did our thing")
 *  DECLINED ("this isn't something with which we want to get involved")
 *  HTTP_mumble ("an error status should be reported")
 */

/* If access is ok, then we decline, otherwise we return FORBIDDEN, or
   OK with an error message. */
				/* forward decl */
static int auth_cert_forbid(request_rec *r, auth_cert_config *cfg) {
  return HTTP_FORBIDDEN;
}

static int auth_cert_forbid_handler(request_rec *r, auth_cert_config *cfg) {
  r->content_type = "text/html";
  ap_soft_timeout("mod_auth_cert unauthorized response",r);
  ap_send_http_header(r);
  /*
   * If we're only supposed to send header information (HEAD request), we're
   * already there.
   */
  if (r->header_only) {
    ap_kill_timeout (r);
    return HTTP_FORBIDDEN;
  }
  /*
   * Now send our actual output.  Since we tagged this as being
   * "text/html", we need to embed any HTML.
   */
  ap_rputs ("<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 3.2//EN\">\n", r);
  ap_rputs ("<HTML>\n", r);
  ap_rputs (" <HEAD>\n", r);
  ap_rputs (" <TITLE>Unauthorized or unauthenticated\n", r);
  ap_rputs (" </TITLE>\n", r);
  ap_rputs (" </HEAD>\n", r);
  ap_rputs (" <BODY>\n", r);
  ap_rputs (" <H1>Unauthorized or unauthenticated\n", r);
  ap_rputs (" </h1>\n", r);
  ap_rprintf (r,"You are unauthorized to access this URL: %s<p>\n",r->uri ? r->uri : "could not get URL");
  if (!r->connection->client->szClientX509) {
    ap_rputs ("You need to have a client-side certificate<p>\n",r);
  } else {
      ap_rprintf (r,"User<br>\n'%s'<br>\nnot authorized<p>\n",r->connection->client->szClientX509);
  }
  ap_rputs (" </BODY>\n", r);
  ap_rputs ("</HTML>\n", r);
  ap_kill_timeout(r);
  return H_OK;
}

/* I don't think this is ever called. 
 * The code was obsolete so I deleted it. It should be the same
 * as auth_cert_ckaccess -mrt
 */
int auth_cert_handler(request_rec *r) {
  auth_cert_config   *cfg;

  cfg = our_dconfig (r);
  trace_add (r->server, cfg, "auth_cert_handler()");
  show_env(r);
  return auth_cert_forbid(r,cfg);

}

/*--------------------------------------------------------------------------*/
/*									    */
/* Now let's declare routines for each of the callback phase in order.	    */
/* (That's the order in which they're listed in the callback list, *not	    */
/* the order in which the server calls them!  See the command_rec	    */
/* declaration near the bottom of this file.)  Note that these may be	    */
/* called for situations that don't relate primarily to our function - in   */
/* other words, the fixup handler shouldn't assume that the request has	    */
/* to do with "example" stuff.	    					    */
/*									    */
/* With the exception of the content handler, all of our routines will be   */
/* called for each request, unless an earlier handler from another module   */
/* aborted the sequence.						    */
/*									    */
/* Handlers that are declared as "int" can return the following:	    */
/*									    */
/*  OK		Handler accepted the request and did its thing with it.	    */
/*  DECLINED	Handler took no action.					    */
/*  HTTP_mumble	Handler looked at request and found it wanting.		    */
/*									    */
/* What the server does after calling a module handler depends upon the	    */
/* handler's return value.  In all cases, if the handler returns	    */
/* DECLINED, the server will continue to the next module with an handler    */
/* for the current phase.  However, if the handler return a non-OK,	    */
/* non-DECLINED status, the server aborts the request right there.  If	    */
/* the handler returns OK, the server's next action is phase-specific;	    */
/* see the individual handler comments below for details.		    */
/*									    */
/*--------------------------------------------------------------------------*/
/* 
 * This function is called during server initialisation.  Any information
 * that needs to be recorded must be in static cells, since there's no
 * configuration record.
 *
 * There is no return value.
 */

/*
 * All our module-initialiser does is add its trace to the log.
 */
static void example_init
	(server_rec *s, pool *p) {

    char    *note;
    char    *sname = s->server_hostname;

    /*
     * If we haven't already allocated our module-private pool, do so now.
     */
    if (example_pool == NULL) {
	example_pool = ap_make_sub_pool (NULL);
    };
    /*
     * The arbitrary text we add to our trace entry indicates for which server
     * we're being called.
     */
    sname = (sname != NULL) ? sname : "";
    note = ap_pstrcat (p, "example_init(", sname, ")", NULL);
    trace_add (s, NULL, note);
}

/*
 * This function gets called to create up a per-directory configuration
 * record.  This will be called for the "default" server environment, and for
 * each directory for which the parser finds any of our directives applicable.
 * If a directory doesn't have any of our directives involved (i.e., they
 * aren't in the .htaccess file, or a <Location>, <Directory>, or related
 * block), this routine will *not* be called - the configuration for the
 * closest ancestor is used.
 *
 * The return value is a pointer to the created module-specific
 * structure.
 */
static void *example_dir_create
	(pool *p, char *dirspec) {

    auth_cert_config
	    *cfg;
    char    *dname = dirspec;

    /*
     * Allocate the space for our record from the pool supplied.
     */
    cfg = (auth_cert_config *) ap_pcalloc (p, sizeof(auth_cert_config));
    /*
     * Now fill in the defaults.  If there are any `parent' configuration
     * records, they'll get merged as part of a separate callback.
     */
    cfg->local = 0;
    cfg->congenital = 0;
    cfg->cmode = CONFIG_MODE_DIRECTORY;
				/* NULL use condition */
    /*
     * Finally, add our trace to the callback list.
     */
    dname = (dname != NULL) ? dname : "";
    cfg->loc = ap_pstrcat (p, "DIR(", dname, ")", NULL);
    trace_add (NULL, cfg, "example_dir_create()");
    return (void *) cfg;
}

/*
 * This function gets called to merge two per-directory configuration
 * records.  This is typically done to cope with things like .htaccess files
 * or <Location> directives for directories that are beneath one for which a
 * configuration record was already created.  The routine has the
 * responsibility of creating a new record and merging the contents of the
 * other two into it appropriately.  If the module doesn't declare a merge
 * routine, the record for the closest ancestor location (that has one) is
 * used exclusively.
 *
 * The routine MUST NOT modify any of its arguments!
 *
 * The return value is a pointer to the created module-specific structure
 * containing the merged values.
 */
static void *example_dir_merge
	(pool *p, void *parent_conf, void *newloc_conf) {

    auth_cert_config
	    *merged_config =
		(auth_cert_config *) ap_pcalloc (p, sizeof(auth_cert_config));
    auth_cert_config
	    *pconf = (auth_cert_config *) parent_conf;
    auth_cert_config
	    *nconf = (auth_cert_config *) newloc_conf;
    char    *note;

    /*
     * Some things get copied directly from the more-specific record, rather
     * than getting merged.
     */
    merged_config->local = nconf->local;
    merged_config->loc = ap_pstrdup (p, nconf->loc);

     if(nconf->accessfile.rep()) {
         merged_config->accessfile = nconf->accessfile;
	 pconf->accessfile = nconf->accessfile;
        } else {
         merged_config->accessfile = pconf->accessfile;
        }

    /* directory use condition overrides*/

    /*
     * Others, like the setting of the `congenital' flag, get ORed in.  The
     * setting of that particular flag, for instance, is TRUE if it was ever
     * true anywhere in the upstream configuration.
     */
    merged_config->congenital = (pconf->congenital | pconf->local);
    /*
     * If we're merging records for two different types of environment (server
     * and directory), mark the new record appropriately.  Otherwise, inherit
     * the current value.
     */
    merged_config->cmode =
	(pconf->cmode == nconf->cmode) ? pconf->cmode : CONFIG_MODE_COMBO;
    /*
     * Now just record our being called in the trace list.  Include the
     * locations we were asked to merge.
     */
    note = ap_pstrcat
	    (
		p,
		"example_dir_merge(\"",
		pconf->loc,
		"\",\"",
		nconf->loc,
		"\")",
		NULL
	    );
    trace_add (NULL, merged_config, note);
    return (void *) merged_config;
}

/*
 * This function gets called to create a per-server configuration
 * record.  It will always be called for the "default" server.
 *
 * The return value is a pointer to the created module-specific
 * structure.
 */
static void *example_server_create
	(pool *p, server_rec *s) {

    auth_cert_config
	    *cfg;
    char    *sname = s->server_hostname;

    /*
     * As with the example_dir_create() reoutine, we allocate and fill in an
     * empty record.
     */
    cfg = (auth_cert_config *) ap_pcalloc (p, sizeof(auth_cert_config));
    cfg->local = 0;
    cfg->congenital = 0;
    cfg->cmode = CONFIG_MODE_SERVER;
    /*
     * Note that we were called in the trace list.
     */
    sname = (sname != NULL) ? sname : "";
    cfg->loc = ap_pstrcat (p, "SVR(", sname, ")", NULL);
    trace_add (s, cfg, "example_server_create()");
    return (void *) cfg;
}

/*
 * This function gets called to merge two per-server configuration
 * records.  This is typically done to cope with things like virtual hosts and
 * the default server configuration  The routine has the responsibility of
 * creating a new record and merging the contents of the other two into it
 * appropriately.  If the module doesn't declare a merge routine, the more
 * specific existing record is used exclusively.
 *
 * The routine MUST NOT modify any of its arguments!
 *
 * The return value is a pointer to the created module-specific structure
 * containing the merged values.
 */
static void *example_server_merge
	(pool *p, void *server1_conf, void *server2_conf) {

	// I don't think this routine is ever called at all

    auth_cert_config
	    *merged_config =
		(auth_cert_config *) ap_pcalloc (p, sizeof(auth_cert_config));
    auth_cert_config
	    *s1conf = (auth_cert_config *) server1_conf;
    auth_cert_config
	    *s2conf = (auth_cert_config *) server2_conf;
    char    *note;

    /*
     * Our inheritance rules are our own, and part of our module's semantics.
     * Basically, just note whence we came.
     */
    merged_config->cmode =
	(s1conf->cmode == s2conf->cmode) ? s1conf->cmode : CONFIG_MODE_COMBO;
    merged_config->local = s2conf->local;
    merged_config->congenital = (s1conf->congenital | s1conf->local);
    merged_config->loc = ap_pstrdup (p, s2conf->loc);
    /*
     * Trace our call, including what we were asked to merge.
     */
    note = ap_pstrcat
	    (
		p,
		"example_server_merge(\"",
		s1conf->loc,
		"\",\"",
		s2conf->loc,
		"\")",
		NULL
	    );
    trace_add (NULL, merged_config, note);
    return (void *) merged_config;
}

/*
 * This routine gives our module an opportunity to translate the URI into an
 * actual filename.  If we don't do anything special, the server's default
 * rules (Alias directives and the like) will continue to be followed.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, no
 * further modules are called for this phase.
 */
static int example_xlate
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    /*
     * We don't actually *do* anything here, except note the fact that we were
     * called.
     */
    cfg->loc = ap_pstrcat (r->pool, "URL(", r->uri, ")", "FILE(", r->filename ? r->filename : "", ")", NULL);
    trace_add (r->server, cfg, "example_xlate()");
    return DECLINED;
}

/*
 * This routine is called to check the authentication information sent with
 * the request (such as looking up the user in a database and verifying that
 * the [encrypted] password sent matches the one in the database).
 *
 * The return value is OK, DECLINED, or some HTTP_mumble error (typically
 * HTTP_UNAUTHORIZED).  If we return OK, no other modules are given a chance
 * at the request during this phase.
 */
static int example_ckuser
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    /*
     * Don't do anything except log the call.
     */
    trace_add (r->server, cfg, "example_ckuser()");
    return DECLINED;
}

/*
 * This routine is called to check to see if the resource being requested
 * requires authorisation.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, no
 * other modules are called during this phase.
 *
 * If *all* modules return DECLINED, the request is aborted with a server
 * error.
 */
static int example_ckauth
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    /*
     * Log the call and return OK, or access will be denied (even though we
     * didn't actually do anything).
     */
    cfg->loc = ap_pstrcat (r->pool, "URL(", r->uri, ")", "FILE(", r->filename ? r->filename : "", ")", NULL);
    trace_add (r->server, cfg, "example_ckauth()");
    return H_OK;
}


/*
 * This routine is called to check for any module-specific restrictions placed
 * upon the requested resource.  (See the mod_access module for an example.)
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  All modules with an
 * handler for this phase are called regardless of whether their predecessors
 * return OK or DECLINED.  The first one to return any other status, however,
 * will abort the sequence (and the request) as usual.
 */
int auth_cert_ckaccess
(request_rec *r) {

  auth_cert_config
    *cfg;

  cfg = our_dconfig (r);
  show_env(r);
  cfg->loc = ap_pstrcat(r->pool,"URL(",r->uri,")",
		     "USER(",r->connection ? (r->connection->user ? r->connection->user : "no authentication" ) : "no connection", ") ",
		     "X509.DN(",
		     r->connection ? (r->connection->client ?
				      (r->connection->client->szClientX509 ? r->connection->client->szClientX509 : "no x509") :
				      "no client BUFF") :
		     "no connection", ") ",
		     "X509.cert(",
		     r->connection ? (r->connection->client ?
				      (r->connection->client->ssl ? "present" : "not present") :
				      "no client BUFF") :
		     "no connection", ") ",
		     "cfg.local=",cfg->local ? "local " : " ",
		     "cfg.congenital=",cfg->congenital ? "congenital " : " ",
		     NULL);
  trace_add (r->server, cfg, "auth_cert_ckaccess()");
    if (SSL_get_session(r->connection->client->ssl)) {
      /* XXX: insert use condition rule check here */
      Certificate_X509 x(SSL_get_session(r->connection->client->ssl)->peer);
      BasicString name;
      int status;
      if (x.get_subject_name(&name) == Certificate_X509::OK) {

	/*
	 * We have the relative uri and the absolute pathname of
         * the resource. Akenti wants the pathname of the file relative
	 * to a rootdirectory. 
	 * If r->uri starts with  '~'  then the localroot is the user's
         *   home directory and must be input to auth_cert_ckaccess
	 * If r->uri starts with '/', then the directory is relative to
         *   the documentroot specified in http.conf
	 */

	BasicString_util bs;
	BasicString localdocroot(""), accesspath("");
	BasicString slash("/");
	if(strstr(r->uri, "~") != NULL) { // user's directory
		// now get the users name
		BasicString t1,usr,t3,t4;
		bs.after(BasicString(r->uri),BasicString("~"),&t1);
		bs.before(t1,slash,&usr); // this gives user's name
		bs.after(t1,slash,&t4);  //  directory relative to ~user
		accesspath = t4;
		// now extract the localdocroot from r->filename
		bs.before(BasicString(r->filename),usr,&t3);
		localdocroot += t3;
		localdocroot += usr;
		bs.after(BasicString(r->filename),usr,&t1);
		bs.before(t1,t4,&t3);  // should be UserDir eg public_html
		localdocroot += t3;
	 } else { // directory relative to the documentroot set in httpd.conf
		BasicString t1;
		BasicString start(r->filename);
		accesspath = r->uri;
		/*
		 * A problem occurred when cgi script was executed with
		 * arguments following the script as a directory for ex.
		 * printenv.cgi/foo (where /foo is an argument), then
		 * the r->path_info is not just a slash (which happens
		 * when you access a html directory, instead of a .html file)
		 * but it includes all the arguments. So add that to the 
		 * filename too. 
		 */
		
		if (bs.startsWith(BasicString(r->path_info), slash) ==
						BasicString_util::OK) {
			start += BasicString(r->path_info);
		 }
		if (BasicString(r->uri) == slash) {
		   	localdocroot = start; 
		 } else {
		   	bs.before(start, BasicString(r->uri), &t1);
			localdocroot += t1;
			localdocroot += slash;
		 }
	 }

	/*
	 * If you see the comment above, to keep the accesspath in
	 * synchronization with the uri requested, we need to add
	 * whatever is in the path_info of request_rec and not leave
	 * with just adding the  slash. 
	 */
	/* if (bs.startsWith(BasicString(r->path_info), slash) ==
					BasicString_util::OK) {
		accesspath += BasicString(r->path_info);
	 } */
	table *e = r->subprocess_env;
	BasicString  res;
	/*
	 * The following env. variable is
	 * not necessary.
	 */
	ap_table_set(e, "USERID", name.rep());
	AuthCertModule accesscontrol;
	BasicString allowed;
	BasicString static protocol("");
	BasicString remotehost (r->connection->remote_ip);

	BasicString serverhost(r->server->server_hostname);
	if(r->protocol != NULL && !strcmp(r->protocol,"HTTP/1.0")) {
		protocol = BasicString("http");
	  }
	BasicString reqResource("");
	reqResource += protocol;
	reqResource += BasicString("://");
	reqResource += serverhost;
	reqResource += BasicString(r->uri);

	auth_cert_config  *scfg = our_sconfig(r->server);
	if (accesscontrol.auth_cert_ckaccess(name, reqResource, accesspath, 
	        remotehost,
		*(scfg->config),localdocroot,&allowed)
				 		== AuthCertModule::OK) {
		//cout << "Allowed actions :" << allowed.rep() << endl;
	        ap_table_set(e, "PRIVILEGE", allowed.rep());
		ap_table_set(e,"AKENTI_CONF",AkentiConf::conf_file_name.rep());

		Search srch;
		/*
		 * Now look for the possibility of the requested
		 * document to be a cgi script.
		 */
		if (srch.compareString(r->handler,"cgi-script") ==
						   Search::OK ) {
		  // It should be a cgi-script
		  BasicString match;
		  if (srch.findstring( BasicString("execute"),allowed,
					    &match) == Search::OK) {
			return H_OK;
		   }
	        } else {
		    return H_OK;
		}
	  } else { 
		return auth_cert_forbid(r,cfg);
	  }
      } else {
	return auth_cert_forbid(r,cfg);
      }
    } else {
      return auth_cert_forbid(r,cfg);
    }
    return auth_cert_forbid(r,cfg);
}

/*
 * This routine is called to determine and/or set the various document type
 * information bits, like Content-type (via r->content_type), language, et
 * cetera.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, no
 * further modules are given a chance at the request for this phase.
 */
static int example_typer
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    /*
     * Log the call, but don't do anything else - and report truthfully that
     * we didn't do anything.
     */
    cfg->loc = ap_pstrcat (r->pool, "URL(", r->uri, ")", "FILE(", r->filename ? r->filename : "", ")", NULL);
    trace_add (r->server, cfg, "example_typer()");
    return DECLINED;
}

/*
 * This routine is called to perform any module-specific fixing of header
 * fields, et cetera.  It is invoked just before any content-handler.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, the
 * server will still call any remaining modules with an handler for this
 * phase.
 */
static int example_fixer
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    /*
     * Log the call and exit.
     */
    trace_add (r->server, cfg, "example_fixer()");
    return H_OK;
}

/*
 * This routine is called to perform any module-specific logging activities
 * over and above the normal server things.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, any
 * remaining modules with an handler for this phase will still be called.
 */
static int example_logger
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    trace_add (r->server, cfg, "example_logger()");
    return DECLINED;
}

/*
 * This routine is called to give the module a chance to look at the request
 * headers and take any appropriate specific actions early in the processing
 * sequence.
 *
 * The return value is OK, DECLINED, or HTTP_mumble.  If we return OK, any
 * remaining modules with handlers for this phase will still be called.
 */
static int example_hparser
	(request_rec *r) {

    auth_cert_config
	    *cfg;

    cfg = our_dconfig (r);
    trace_add (r->server, cfg, "example_hparser()");
    return DECLINED;
}

/*--------------------------------------------------------------------------*/
/*									    */
/* All of the routines have been declared now.  Here's the list of	    */
/* directives specific to our module, and information about where they	    */
/* may appear and how the command parser should pass them to us for	    */
/* processing.  Note that care must be taken to ensure that there are NO    */
/* collisions of directive names between modules.			    */
/*									    */
/*--------------------------------------------------------------------------*/
/* 
 * List of directives specific to our module.
 */
command_rec auth_cert_commands[] = {
    {
        "AuthenticateWithCert",                 /* directive name */
        (const char *(*)())cmd_auth_cert, /* action routine for directive */
        NULL,                           /* argument to include in call */
        OR_LIMIT,                       /* where available */
        NO_ARGS,                        /* arguments */
        "Authentication directive - no arguments"
                                        /* directive description */
    },
        /******** SYNTAX : to be removed in the future ********/
    {
      "CertificateDirectory",   /* where do certificates come from */
        (const char *(*)())cmd_auth_certdir, /* action routine for directive */
        NULL,                           /* argument to include in call */
        OR_LIMIT,                       /* where available */
        TAKE3,                  /* arguments */
        "CertificateDirectory public|trusted web|file hostname|directory"
                                        /* directive description */
    },
    {
      "UseCondRequired",
      (const char *(*)())cmd_auth_usecond,
      NULL,
      OR_LIMIT,                 /* if LIMIT override is allowed */
      ITERATE,                    /* parse the arguments ourselves */
      "UseCondRequired url"
    },
    {
      "UserIdCertificateAuthority",
      (const char *(*)())cmd_auth_certauthority,
      NULL,
      OR_LIMIT,
      TAKE2,                    /* distinguished name and the certificate */
      "UserIdCertificateAuthority distinguished_name certificate"
                                /* description */
    },
    {
      "UseConditionCAandIssuer",
      (const char *(*)())cmd_auth_usecondissuer,
      NULL,
      OR_LIMIT,
      ITERATE,                  /* distinguished names of CA and Issuer */
      "UserConditionCAandIssuer dnca dnissuer" /* directive description */
    },
    {
      "AkentiConf",
      (const char *(*)())cmd_auth_akenticonf,
      NULL,
      RSRC_CONF,
      TAKE1,                  /* distinguished names of CA and Issuer */
      "AkentiConf AkentiConf_pathname" /* directive description */
    },
        /******** SYNTAX : to be provided in the future ********/
    {
      "UseConditionDirectory", /* where to search for use condition certificates
 */
      (const char *(*)())cmd_auth_useconddir, /* action routine for directive */
      NULL,
      OR_LIMIT,
      TAKE3,
      "UseConditionDirectory trusted web hostname" /* directive description */
     },
    { NULL }
};

/*--------------------------------------------------------------------------*/
/*									    */
/* Now the list of content handlers available from this module.		    */
/*									    */
/*--------------------------------------------------------------------------*/
/* 
 * List of content handlers our module supplies.  Each handler is defined by
 * two parts: a name by which it can be referenced (such as by
 * {Add,Set}Handler), and the actual routine name.  The list is terminated by
 * a NULL block, since it can be of variable length.
 *
 * Note that content-handlers are invoked on a most-specific to least-specific
 * basis; that is, a handler that is declared for "text/plain" will be
 * invoked before one that was declared for "text / *".  Note also that
 * if a content-handler returns anything except DECLINED, no other
 * content-handlers will be called.
 */
handler_rec auth_cert_handlers[] = {
    { "auth-cert-handler", auth_cert_handler },
    { NULL }
};

/*--------------------------------------------------------------------------*/
/*									    */
/* Finally, the list of callback routines and data structures that	    */
/* provide the hooks into our module from the other parts of the server.    */
/*									    */
/*--------------------------------------------------------------------------*/
/* 
 * Module definition for configuration.  If a particular callback is not
 * needed, replace its routine name below with the word NULL.
 *
 * The number in brackets indicates the order in which the routine is called
 * during request processing.  Note that not all routines are necessarily
 * called (such as if a resource doesn't have access restrictions).
 */
module auth_cert_module = {
    STANDARD_MODULE_STUFF,
    example_init,		/* initializer */
    example_dir_create,		/* per-directory config creater */
    example_dir_merge,		/* dir config merger - default is to override */
    example_server_create,	/* server config creator */
    example_server_merge,	/* server config merger */
    auth_cert_commands,		/* command table */
    auth_cert_handlers,		/* [6] list of handlers */
    example_xlate,		/* [1] filename-to-URI translation */
    example_ckuser,		/* [4] check/validate HTTP user_id */
    example_ckauth,		/* [5] check HTTP user_id is valid *here* */
    auth_cert_ckaccess,		/* [3] check access by host address, etc. */
    example_typer,		/* [6] MIME type checker/setter */
    example_fixer,		/* [7] fixups */
    example_logger,		/* [9] logger */
    example_hparser		/* [2] header parser */
};

/*--------------------------------------------------------------------------*/
/*									    */
/* These routines are strictly internal to this module, and support its	    */
/* operation.  They are not referenced by any external portion of the	    */
/* server.								    */
/*									    */
/*--------------------------------------------------------------------------*/

/*
 * Locate our directory configuration record for the current request.
 */
static auth_cert_config *our_dconfig
	(request_rec *r) {

    return (auth_cert_config *) ap_get_module_config
				(
				    r->per_dir_config,
				    &auth_cert_module
				);
}

/*
 * Locate our server configuration record for the specified server.
 */
static auth_cert_config *our_sconfig
	(server_rec *s) {

    return (auth_cert_config *) ap_get_module_config
				(
				    s->module_config,
				    &auth_cert_module
				);
}
